from typing import Callable, Literal, Tuple, Union

import numpy as np

from bermuda.triangle import Triangle

__all__ = ["moment_match"]


def _sort_x_on_y_rank(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    # more efficient than y.argsort().argsort(), sorting only done once
    sort_y = y.argsort()
    rank_y = np.empty_like(sort_y)
    rank_y[sort_y] = np.arange(len(y))

    return np.array(sorted(x))[rank_y]


def _get_sample_moments(samples: np.ndarray) -> Tuple[float, float, int]:
    return (np.mean(samples), np.var(samples), len(samples))


def _sample_normal_dist(mu: float, sigma2: float, n: int) -> np.ndarray:
    return np.random.normal(loc=mu, scale=np.sqrt(sigma2), size=n)


def _sample_lognormal_dist(mu: float, sigma2: float, n: int) -> np.ndarray:
    mean = (mu**2) / np.sqrt((mu**2) + sigma2)
    var = np.log(1 + sigma2 / (mu**2))
    return np.random.lognormal(mean=np.log(mean), sigma=np.sqrt(var), size=n)


def _sample_gamma_dist(mu: float, sigma2: float, n: int) -> np.ndarray:
    shape = mu**2 / sigma2
    rate = mu / sigma2
    return np.random.gamma(shape=shape, scale=1 / rate, size=n)


SAMPLE_DIST_dict = {
    "normal": _sample_normal_dist,
    "lognormal": _sample_lognormal_dist,
    "gamma": _sample_gamma_dist,
}


def _generate_samples(
    samples: Union[np.ndarray, float, int],
    sample_fn: Callable[[float, float, int], np.ndarray],
) -> Union[np.ndarray, float, int]:
    if type(samples) is not np.ndarray:
        return samples

    moments = _get_sample_moments(samples)

    return np.array(_sort_x_on_y_rank(sample_fn(*moments), samples))


def moment_match(
    triangle: Triangle,
    field_names: list[str],
    distribution: Literal["normal", "lognormal", "gamma"],
) -> Triangle:
    """Use method of moments matching to convert samples in a triangle from one distribution to another.

    Args:
        triangle: bermuda.triangle.Triangle with fields that need converted.
        field_names: List of field names to apply conversion to. Typical use cases include "reported_loss", "paid_loss",
        "incurred_loss", etc.
        distribution: Distribution to convert sample to. The empirical sample mean and variance are matched to moments of
        `distribution`, which is then sampled from.

    Raises:
        KeyError: if supplied `field_names` are not all contained in the input triangle.

    Returns:
        Triangle: bermuda.triangle.Triangle with `field_names` samples in each cell replaced with samples from `distribution`. If a
        field contains only scalar values in each cell, the scalar values are returned (no moment matching is done). Otherwise, if the
        field contains arrays of samples, moment matched samples will be returned (all with the same rank order as the original
        samples).
    """

    field_name_diff = list(set(field_names) - set(triangle.fields))
    if len(field_name_diff) > 0:
        raise KeyError(
            f"The following supplied field_names do not exist in `triangle`: {field_name_diff}"
        )

    for field in field_names:
        triangle = triangle.derive_fields(
            **{
                field: lambda ob: _generate_samples(
                    ob[field], SAMPLE_DIST_dict[distribution]
                )
            }
        )

    return triangle



from __future__ import annotations

import datetime
import re
import warnings
from collections import abc, defaultdict
from functools import cached_property
from typing import Any, Callable, Sequence

import numpy as np
import pandas as pd
import toolz as tlz

import altair as alt

from .base import Cell, IncrementalCell, Metadata, common_metadata, metadata_diff
from .date_utils import dev_lag_months, eval_date_resolution, period_resolution
from .errors import DuplicateCellWarning, TriangleEmptyError, TriangleError

__all__ = [
    "Triangle",
    "TriangleSlice",
]


class Triangle(abc.Set):
    """The main triangle class.

    The Triangle class inherits from collections.abc.Set.
    It holds a sequence of Cell objects, which are automatically
    sorted on class instance creation.

    The Triangle class has a number of methods, properties
    and static methods for triangle manipulation, visualization,
    and I/O. Consult the documentation for more details.

    Attributes:
        cells: the sequence of Cell objects.
    """

    def __init__(self, cells: Sequence[Cell]) -> None:
        if any(not isinstance(cell, Cell) for cell in cells):
            raise TriangleError("Bermuda triangles can only hold `Cell`s")

        # check for cell type consistency
        if not (
            all(cell.__class__.__name__ == "Cell" for cell in list(cells))
            or all(cell.__class__.__name__ == "CumulativeCell" for cell in list(cells))
            or all(cell.__class__.__name__ == "IncrementalCell" for cell in list(cells))
        ):
            raise TriangleError("Triangle cells must have consistent type.")

        try:
            self._cells = sorted(list(cells))
        except Exception as e:
            raise TriangleError(
                "Triangle must be constructed with some sequence of Cells"
            ) from e

        # The cells are already sorted by metadata and coordinates, so checking
        # for duplicates is as easy as checking for consecutive duplicates
        for before, after in zip(self._cells[:-1], self._cells[1:]):
            if (
                before.metadata == after.metadata
                and before.coordinates == after.coordinates
            ):
                warnings.warn(
                    "This triangle has multiple cells with the same metadata and coordinates "
                    f"at period_start {before.period_start} "
                    f"and evaluation_date {before.evaluation_date}",
                    DuplicateCellWarning,
                )
                break
            if isinstance(self._cells[0], IncrementalCell):
                if (
                    before.metadata == after.metadata
                    and before.period == after.period
                    and before.evaluation_date == after.evaluation_date
                    and before.prev_evaluation_date != after.prev_evaluation_date
                ):
                    warnings.warn(
                        "This incremental triangle has multiple cells with the same metadata "
                        f"and coordinates at period_start {before.period_start} "
                        f"and evaluation_date {before.evaluation_date} "
                        "with different prev_evaluation_dates",
                        DuplicateCellWarning,
                    )
                    break

    def __hash__(self):
        return hash(tuple(self._cells))

    def __eq__(self, other) -> bool:
        return all([cell1 == cell2 for cell1, cell2 in zip(self.cells, other.cells)])

    def __len__(self) -> int:
        return len(self._cells)

    def __iter__(self):
        return iter(self._cells)

    def __contains__(self, cell):
        return cell in self._cells

    def _repr_html_(self):
        if len(self) == 0:
            table = [
                "<table>",
                "  <thead>",
                "    <tr><th> <h4><b>Empty Triangle</b></h4> </th></tr>",
                "  </thead>",
                "  <tbody>",
                "    <tr><th> Number of slices: </th><td> 0 </td></tr>",
                "    <tr><th> Number of cells: </th><td> 0 </td></tr>",
                "  </tbody>",
                "</table>",
            ]
            return "\n".join(table)

        metadata_rows = []
        empty_column = "<td>   </td>"
        for name, value in self.common_metadata.as_flat_dict().items():
            if value is not None:
                metadata_rows.append(
                    f"<tr>{empty_column}<th>{name}</th> <td> {value} </td></tr>"
                )

        experience_gap_rows = []
        if self.experience_gaps:
            experience_gap_rows.append(
                '<tr><th colspan=2 style="text-align:center"> Experience Gaps </th></tr>'
            )
            for gap in self.experience_gaps:
                experience_gap_rows.append(
                    f"<tr><td></td><td>{gap[0]}/{gap[1]}</td></tr>"
                )

        standard_field_rows = []
        sparse_field_rows = []
        for name, count in self.field_cell_counts.items():
            if count == len(self):
                standard_field_rows.append(f"<tr>{empty_column}<td>{name}</td></tr>")
            else:
                hit_rate = 100 * count / len(self)
                sparse_field_rows.append(
                    f"<tr>{empty_column}<td>{name} ({hit_rate:.1f}% coverage)</td></tr>"
                )
        if standard_field_rows:
            standard_field_rows.insert(
                0, '<tr><th colspan=2 style="text-align:center"> Fields: </th></tr>'
            )
        if sparse_field_rows:
            sparse_field_rows.insert(
                0,
                '<tr><th colspan=2 style="text-align:center"> Optional Fields: </th></tr>',
            )

        if not self.is_slicewise_disjoint:
            category = "Non-Disjoint"
        elif not self.is_semi_regular():
            category = "Slicewise Disjoint"
        elif not self.is_regular():
            category = "Semi-Regular"
        else:
            category = "Regular"

        experience_range = f"{self.periods[0][0]}/{self.periods[-1][1]}"
        eval_date_range = f"{self.evaluation_dates[0]}/{self.evaluation_dates[-1]}"
        dev_lag_range = f"{self.dev_lags()[0]} - {self.dev_lags()[-1]}"
        tri_heading = (
            "Incremental Triangle"
            if isinstance(self._cells[0], IncrementalCell)
            else "Cumulative Triangle"
        )

        table = [
            f"<table>  <thead>    <tr><th> <h4><b>{tri_heading}</b></h4> </th></tr>",
            "  </thead>",
            "  <tbody>",
            f"    <tr><th> Number of slices: </th><td> {len(self.slices)} </td></tr>",
            f"    <tr><th> Number of cells: </th><td> {len(self)} </td></tr>",
            f"    <tr><th> Triangle category: </th><td> {category} </td></tr>",
            f"    <tr><th> Experience range: </th><td> {experience_range} </td></tr>",
            f"    <tr><th> Experience resolution: </th><td> {self.period_resolution} </td></tr>",
            f"    <tr><th> Evaluation range: </th><td> {eval_date_range} </td></tr>",
            f"    <tr><th> Evaluation resolution: </th><td> {self.eval_date_resolution} </td></tr>",
            f"    <tr><th> Dev Lag range: </th><td> {dev_lag_range} months </td></tr>",
            *experience_gap_rows,
            *standard_field_rows,
            *sparse_field_rows,
            '    <tr><th colspan=2 style="text-align:center"> Common Metadata: </th></tr>',
            *metadata_rows,
            "  </tbody>",
            "</table>",
        ]
        return "\n".join(table)

    def __repr__(self):
        return strip_html_tags(
            "\n".join(line.strip() for line in self._repr_html_().splitlines())
        )

    def __getitem__(self, index):
        """Triangle allows indexing by integer to directly retrieve a cell, or by
        (period, evaluation_date, metadata) indices. Period is indexed by period
        start. All indices can be passed as slices. Use `:` to indicate an empty
        slice selecting over all available cells. If a slice is passed to any index
        a Triangle object will be returned. If no slices are passed then the single
        Cell at the index will be returned."""
        # Direct Cell Indexing
        if isinstance(index, int):
            return self._cells[index]
        if isinstance(index, slice):
            return Triangle(self._cells[index])

        # Index by period, evaluation_date, metadata
        if len(index) == 3:
            period_slice, evaluation_slice, metadata = index
        else:
            raise ValueError("Must pass three indices to Triangle, or a single integer")

        if metadata and metadata != slice(None, None, None):
            filtered = self.filter(lambda cell: cell.metadata == metadata)
        else:
            filtered = self

        if isinstance(period_slice, slice):
            period_start = period_slice.start
            period_end = period_slice.stop
        elif isinstance(period_slice, datetime.date):
            period_start, period_end = period_slice, period_slice
        else:
            raise ValueError(
                f"period_slice must be date or slice, received {period_slice}"
            )

        # Clip cuts off based on period_end, we use filter to keep based on period_start
        if not period_start:
            period_start = datetime.date.min
        if not period_end:
            period_end = datetime.date.max
        filtered = filtered.filter(
            lambda cell: period_start <= cell.period_start <= period_end
        )

        if isinstance(evaluation_slice, slice):
            evaluation_start = evaluation_slice.start
            evaluation_end = evaluation_slice.stop
        elif isinstance(evaluation_slice, datetime.date):
            evaluation_start, evaluation_end = evaluation_slice, evaluation_slice
        else:
            raise ValueError(
                f"evaluation_slice must be date, received {evaluation_slice}"
            )

        clipped = filtered.clip(
            min_eval=evaluation_start,
            max_eval=evaluation_end,
        )
        if any(isinstance(ind, slice) for ind in index):
            return clipped
        return clipped._cells[0]

    @property
    def cells(self):
        return self._cells

    @property
    def slices(self) -> dict[Metadata, "Triangle"]:
        return tlz.valmap(
            Triangle, tlz.groupby(lambda cell: cell.metadata, self._cells)
        )

    @cached_property
    def metadata(self) -> list[Metadata]:
        """A sorted list of all unique metadata in the Triangle."""
        return sorted(set([cell.metadata for cell in self._cells]))

    @cached_property
    def eval_date_resolution(self) -> int:
        """The resolution of the evaluation dates in the triangle."""
        return eval_date_resolution(self)

    @cached_property
    def period_resolution(self) -> int:
        """The resolution of the periods in the triangle."""
        return period_resolution(self)

    @cached_property
    def common_metadata(self) -> Metadata:
        """A single metadata element representing the metadata attributes common to all
        Cells in the Triangle."""
        metas = self.metadata
        if len(metas) == 1:
            return metas[0]

        common_meta = metas[0]
        for meta in metas[1:]:
            common_meta = common_metadata(common_meta, meta)
        return common_meta

    @cached_property
    def metadata_differences(self) -> list[Metadata]:
        """A list of all unique metadata in the Triangle, with metadata attributes
        that are common to all Cells in the Triangle removed."""
        return [metadata_diff(self.common_metadata, meta) for meta in self.metadata]

    @cached_property
    def num_samples(self) -> int:
        """The number of samples in a Triangle of stochastic predictions. If the triangle
        has only observed data, `num_samples` is 1. If different cells or different fields
        have different values of `num_samples`, an error is raised."""
        num_samples = None
        for cell in self._cells:
            for val in cell.values.values():
                # Figure out the size of the value
                if isinstance(val, np.ndarray) and val.size > 1:
                    value_size = val.size
                else:
                    value_size = None

                # Compare and update the overall sample size
                if value_size is not None:
                    if num_samples is None:
                        num_samples = value_size
                    elif num_samples != value_size:
                        raise ValueError("The triangle has inconsistent value sizes")

        return 1 if num_samples is None else num_samples

    @property
    def period_rows(self):
        """Group by experience period and return an iterator over 'rows'."""
        grouped = tlz.groupby(lambda cell: cell.period, self._cells)
        for period in self.periods:
            yield (
                period,
                sorted(
                    grouped[period],
                    key=lambda cell: (cell.metadata, cell.evaluation_date),
                ),
            )

    @property
    def slice_period_rows(self):
        """Group by slice and experience period and return an iterator over 'slice-rows'."""
        grouped = tlz.groupby(lambda cell: (cell.metadata, cell.period), self._cells)
        for key, row in sorted(grouped.items()):
            yield key, sorted(row, key=lambda cell: cell.evaluation_date)

    @cached_property
    def periods(self) -> list[tuple[datetime.date, datetime.date]]:
        """A sorted list of unique experience periods (as 2-tuples of start and end dates)."""
        return sorted({cell.period for cell in self._cells})

    @cached_property
    def experience_gaps(self) -> list[tuple[datetime.date, datetime.date]]:
        """A sorted list of missing experience period ranges from the triangle"""
        missing_periods = []
        for (_, current_end), (next_start, _) in zip(
            self.periods[:-1], self.periods[1:]
        ):
            continuous_next_start = current_end + datetime.timedelta(days=1)
            if next_start != continuous_next_start:
                missing_periods += [
                    (continuous_next_start, next_start - datetime.timedelta(days=1))
                ]

        return missing_periods

    def dev_lags(self, unit="month") -> list[float | int | datetime.timedelta]:
        """A sorted list of unique development lags.  unit can be, 'month', 'day' or 'timedelta'.
        See docs for `Observation.dev_lag()` or `Prediction.dev_lag()` for more information.
        """
        return sorted({cell.dev_lag(unit) for cell in self._cells})

    @cached_property
    def evaluation_dates(self) -> list[datetime.date]:
        """A sorted list of unique evaluation dates."""
        return sorted({cell.evaluation_date for cell in self._cells})

    @cached_property
    def evaluation_date(self) -> datetime.date:
        """The latest evaluation date in the triangle."""
        if self.is_empty:
            raise TriangleEmptyError
        else:
            return max(self.evaluation_dates)

    @cached_property
    def fields(self) -> list[str]:
        """A sorted list of unique fields in the triangle set."""
        return sorted({name for cell in self._cells for name in cell.values.keys()})

    @cached_property
    def field_cell_counts(self) -> dict[str, int]:
        """For each triangle field return the number of cells the field appears in."""
        fcc = dict()
        for field in self.fields:
            fcc[field] = sum([field in cell.values.keys() for cell in self.cells])
        return fcc

    @cached_property
    def field_slice_counts(self) -> dict[str, int]:
        """For each triangle field return the number of slices the field appears in."""
        fsc = dict()
        for field in self.fields:
            fsc[field] = sum([field in slc.fields for slc in self.slices.values()])
        return fsc

    def __add__(self, other: Triangle) -> Triangle:
        """Concatenate `self` with `other` (another `Triangle`)."""
        return Triangle(self._cells + other._cells)

    def __radd__(self, other: Triangle) -> Triangle:
        # Enable summing triangle sets using built-in `sum` function.
        if other == 0:
            return self
        else:
            return self.__add__(other)

    @cached_property
    def is_empty(self) -> bool:
        """Is the triangle empty?"""
        return self._cells == []

    @cached_property
    def is_disjoint(self) -> bool:
        """Is the triangle disjoint?

        A triangle is disjoint if there is no overlap between any pair of experience periods.
        """
        if self.is_empty:
            return True

        for (prev_start, prev_end), (next_start, next_end) in zip(
            self.periods[:-1], self.periods[1:]
        ):
            if prev_end >= next_start:
                return False
        return True

    @cached_property
    def is_slicewise_disjoint(self) -> bool:
        """Returns True if each triangle slice is disjoint."""
        for slc in self.slices.values():
            if not slc.is_disjoint:
                return False
        return True

    def is_semi_regular(self, dev_lag_unit="month") -> bool:
        """Is the triangle semi-regular?

        A triangle is semi-regular if it is disjoint and every experience period has the
        same duration."""
        if not self.is_disjoint:
            return False

        if self.is_empty:
            return True

        base_start, base_stop = self.periods[0]

        _unit = dev_lag_unit.lower()
        if "month" in _unit:
            # noinspection PyShadowingNames
            def diff_fn(start, stop):
                return dev_lag_months(start - datetime.timedelta(days=1), stop)

        elif "day" in _unit or _unit == "timedelta":
            # noinspection PyShadowingNames
            def diff_fn(start, stop):
                return start - stop

        else:
            raise ValueError(
                f"Unrecognized dev_lag unit '{dev_lag_unit}', "
                "must be one of 'month', 'day' or 'timedelta'."
            )

        base_length = diff_fn(base_start, base_stop)

        for start, stop in self.periods[1:]:
            if diff_fn(start, stop) != base_length:
                return False

        return True

    def is_regular(self, dev_lag_unit="month") -> bool:
        """Is the triangle regular?

        A triangle is regular if it is semi-regular and the interval between consecutive
        development lags is always a constant."""
        if not self.is_semi_regular(dev_lag_unit):
            return False
        if self.is_empty:
            return True

        dev_lags = self.dev_lags(unit=dev_lag_unit)
        # If there's only one dev lag, no need to check for consistent intervals
        if len(dev_lags) == 1:
            return True

        dev_offset = dev_lags[1] - dev_lags[0]
        for prv, nxt in zip(dev_lags[1:-1], dev_lags[2:]):
            if nxt - prv != dev_offset:
                return False

        return True

    @cached_property
    def has_consistent_currency(self) -> bool:
        """Does every cell in the triangle use the same currency?"""
        return len({cell.metadata.currency for cell in self.cells}) == 1

    @cached_property
    def has_consistent_risk_basis(self) -> bool:
        """Does every cell in the triangle use the same risk basis?"""
        # Risk basis is the third field in the metadata tuple
        return len({cell.metadata.risk_basis for cell in self.cells}) == 1

    @cached_property
    def is_incremental(self) -> bool:
        """Is the traingle incremental?"""
        return isinstance(self.cells[0], IncrementalCell)

    @cached_property
    def is_multi_slice(self) -> bool:
        """Does this triangle have multiple slices?"""
        return len(self.slices) > 1

    def derive_fields(self, **definitions: Callable[[Cell], Any] | Any) -> Triangle:
        """Create a new `Triangle` with additional derived fields defined.

        Args:
            **definitions: New field definitions. Argument names are the names of the
                new fields to be defined, values are either constants or functions that
                accept cells and return field values. Later definitions can depend on
                quantities computed via earlier definitions.
        """
        return Triangle(
            list(map(lambda cell: cell.derive_fields(**definitions), self.cells))
        )

    def derive_metadata(self, **definitions: Callable[[Cell], Any] | Any) -> Triangle:
        """Create a new `Triangle` with additional metadata attributes or fields derived.

        Args:
            **definitions: Values are plain values or functions that accept a `Cell` and
                    return a value. If keys are top-level attributes in metadata, their
                    values will replace the existing values of the top-level attributes.
                    Otherwise, the key-value pair will be added to the `details` field of
                    the metadata.
        """
        return Triangle(
            list(map(lambda cell: cell.derive_metadata(**definitions), self.cells))
        )

    def replace(self, **definitions: Callable[[Cell], Any] | Any) -> Triangle:
        """Replace cell member variables with a constant or via a function.

        Args:
            **definitions: New variable definitions, which must be one of
                `period_start`, `period_end`, `evaluation_date` or `values`.
                Definition values can either be constants or functions of the cell itself.
                For example,
                    triangle.replace(values=dict(reported_loss=1)
                    triangle.replace(values=lambda cell: dict(reported_loss = cell["paid_loss"]))

        """
        return Triangle(list(map(lambda cell: cell.replace(**definitions), self.cells)))

    def remove_static_details(self) -> Triangle:
        """Removes any metadata details and loss_details that are common to all cells in the triangle."""
        if self.is_empty:
            return self
        common_details = self.common_metadata.details.keys()
        common_loss_details = self.common_metadata.loss_details.keys()
        return self.derive_metadata(
            details=lambda meta: {
                k: v for k, v in meta.details.items() if k not in common_details
            },
            loss_details=lambda meta: {
                k: v
                for k, v in meta.loss_details.items()
                if k not in common_loss_details
            },
        )

    @cached_property
    def has_consistent_values_shapes(self) -> bool:
        """Does every field have the same shape in each cell?"""
        shapes = defaultdict(list)
        for cell in self.cells:
            for field, value in cell.values.items():
                shapes[field].append(np.size(value))

        return all(len(set(v)) == 1 for v in shapes.values())

    def select(self, keys: list[str]) -> Triangle:
        """Retain only the designated keys in `values`."""
        return Triangle(list(map(lambda cell: cell.select(keys), self.cells)))

    def clip(
        self,
        *,
        min_eval: datetime.date | None = None,
        max_eval: datetime.date | None = None,
        min_period: datetime.date | None = None,
        max_period: datetime.date | None = None,
        min_dev: datetime.timedelta | int | float | None = None,
        max_dev: datetime.timedelta | int | float | None = None,
        dev_lag_unit: str = "month",
    ) -> Triangle:
        """Create a new `Triangle` that has been clipped/censored as specified.

        Args:
            min_eval: Minimum evaluation date to permit.
            max_eval: Maximum evaluation date to permit.
            min_period: Minimum period start date to permit.
            max_period: Maximum period end date to permit.
            min_dev: Minimum development lag to permit.
            max_dev: Maximum development lag to permit.
            dev_lag_unit: Unit of min_dev or max_dev, required if either is specified.
        """
        cells = self._cells
        if min_eval is not None:
            cells = filter(lambda cell: cell.evaluation_date >= min_eval, cells)
        if max_eval is not None:
            cells = filter(lambda cell: cell.evaluation_date <= max_eval, cells)
        if min_period is not None:
            cells = filter(lambda cell: cell.period_start >= min_period, cells)
        if max_period is not None:
            cells = filter(lambda cell: cell.period_end <= max_period, cells)
        if min_dev is not None:
            cells = filter(lambda cell: cell.dev_lag(dev_lag_unit) >= min_dev, cells)
        if max_dev is not None:
            cells = filter(lambda cell: cell.dev_lag(dev_lag_unit) <= max_dev, cells)

        return Triangle(list(cells))

    def filter(self, predicate: Callable[[Cell], bool]) -> Triangle:
        """Create a new `Triangle` with only the cells that satisfy the predicate.

        Args:
            predicate: A function that accepts a `Cell` and returns a `boolean`.
        """
        return Triangle(list(filter(predicate, self._cells)))

    def extract(self, attribute: Callable[[Cell], Any] | str) -> np.ndarray:
        """Extract cell attributes either via a function or cell field name,
        returning a numpy array of values.

        Args:
            attribute: A function applied to all cells or a cell field.
        """
        if isinstance(attribute, Callable):
            return np.array(list(map(attribute, self._cells)))
        return np.array(list(map(lambda cell: cell.values.get(attribute), self._cells)))

    def to_data_frame(self, dev_lag_unit="month"):
        """Convert a triangle set to a Pandas dataframe."""
        return pd.DataFrame(
            cell.to_record(dev_lag_unit) for cell in self._cells
        ).reset_index(drop=True)

    @property
    def right_edge(self):
        """Create a `Triangle` with only the most recent observation for each period
        within each triangle slice.

        Note that right_edge doesn't guarantee that every cell will have the same evaluation date!
        """
        new_cells = []
        for slc in self.slices.values():
            for period, row in slc.period_rows:
                new_cells.append(row[-1])
        return Triangle(new_cells)

    @property
    def is_right_edge_ragged(self) -> bool:
        """Returns `True` if any slice's right edge encompasses different evaluation dates"""
        for slc in self.slices.values():
            if len(slc.right_edge.evaluation_dates) > 1:
                return True
        return False


class TriangleSlice(Triangle):
    def __init__(self, cells: Sequence[Cell]):
        super().__init__(cells)
        if len(self.slices) > 1:
            raise TriangleError("TriangleSlice cannot have multiple slices.")

    def __getitem__(self, index):
        """TriangleSlice allows indexing by integer to directly retrieve a cell, or by
        (period, evaluation_date) indices. Period is indexed by period
        start. All indices can be passed as slices. Use `:` to indicate an empty
        slice selecting over all available cells. If a slice is passed to any index
        a TriangleSlice object will be returned. If no slices are passed then the single
        Cell at the index will be returned."""
        # Direct Cell Indexing
        if isinstance(index, int):
            return self._cells[index]
        if isinstance(index, slice):
            return TriangleSlice(self._cells[index])

        # Index by period, evaluation_date
        if len(index) == 2:
            period_slice, evaluation_slice = index
        else:
            raise ValueError(
                "Must pass two indices to TriangleSlice, or a single integer"
            )

        if isinstance(period_slice, slice):
            period_start = period_slice.start
            period_end = period_slice.stop
        elif isinstance(period_slice, datetime.date):
            period_start, period_end = period_slice, period_slice
        else:
            raise ValueError(
                f"period_slice must be date or slice, received {period_slice}"
            )

        # Clip cuts off based on period_end, we use filter to keep based on period_start
        if not period_start:
            period_start = datetime.date.min
        if not period_end:
            period_end = datetime.date.max

        filtered = self.filter(
            lambda cell: period_start <= cell.period_start <= period_end
        )

        if isinstance(evaluation_slice, slice):
            evaluation_start = evaluation_slice.start
            evaluation_end = evaluation_slice.stop
        elif isinstance(evaluation_slice, datetime.date):
            evaluation_start, evaluation_end = evaluation_slice, evaluation_slice
        else:
            raise ValueError(
                f"evaluation_slice must be date , received {evaluation_slice}"
            )

        clipped = filtered.clip(
            min_eval=evaluation_start,
            max_eval=evaluation_end,
        )
        if any(isinstance(ind, slice) for ind in index):
            return TriangleSlice(clipped.cells)
        return clipped._cells[0]


def strip_html_tags(data):
    p = re.compile(r"<.*?>")
    return p.sub("", data)



from ..triangle import Triangle, TriangleSlice

__all__ = [
    "slice_to_triangle",
    "triangle_to_slice",
]


def slice_to_triangle(triangle_slice):
    return Triangle(triangle_slice.cells)


def triangle_to_slice(triangle):
    return TriangleSlice(triangle.cells)



import warnings

import numpy as np

from ..base import Cell
from ..date_utils import dev_lag_months, month_to_id
from ..triangle import Triangle
from .basis import to_cumulative
from .merge import period_merge

__all__ = [
    "paid_bs_adjustment",
    "reported_bs_adjustment",
    "weight_geometric_decay",
]


def paid_bs_adjustment(
    triangle: Triangle,
    triangle_ult_claim_counts: Triangle,
) -> Triangle:
    """
    Adjust paid losses in `triangle` using the "paid" Berquist-Sherman adjustment for changes in
    claim settlement rate. The adjustment assumes that the latest diagonal of a triangle will
    reflect current claim settlement rates, so paid losses in other cells are adjusted using
    linear interpolation. If the claim disposal rate for a given cell is less than the selected
    disposal rate for that development lag, losses are adjusted upwards (i.e. using the next cell
    for the same experience period); if the claim disposal rate for a given cell is greater than
    the selected disposal rate for that development lag, losses are adjusted downwards (i.e. using
    the previous cell in the same experience period).

    For more details on Berquist-Sherman, see Chapter 13 of this text:
    https://www.casact.org/sites/default/files/database/studynotes_friedland_estimating.pdf

    Args:
        triangle: The Triangle that will be adjusted. It must include `cwp_claims` and `paid_loss`
            as fields.
        triangle_ult_claim_counts: A right edge Triangle that has the ultimate `reported_claims`
            corresponding to each period in `tri`. The mean of `reported_claims` is used in
            disposal rate calculations, so its value can be either an array or a scalar.

    Returns:
        A Triangle with adjusted `paid_loss` values. Note that closed claim counts remain unchanged
        from the original triangle.
    """
    tri_ult_right_edge = triangle_ult_claim_counts.right_edge

    if triangle.is_incremental:
        triangle = to_cumulative(triangle)

    tri_w_disposal_rates = (
        period_merge(
            triangle,
            tri_ult_right_edge,
            suffix="_ult",
        )
        .derive_fields(
            disposal_rate=lambda ob: ob["cwp_claims"]
            / np.mean(ob["reported_claims_ult"]),
        )
        .select(list(triangle.fields) + ["disposal_rate"])
    )

    # We need to get the cell with the latest period for each dev lag. This accounts
    # for the situation when there are multiple dev lags in the latest period (e.g. 0 and 12 months).
    # Otherwise, we can't get a corresponding disposal rate for all dev lags.
    def _get_max_period_cell_for_dev_lag(c: Cell):
        return tri_w_disposal_rates.filter(
            lambda ob: (ob.dev_lag() == c.dev_lag()) and (ob.metadata == c.metadata)
        )[-1]  # relies on sorting of cells by increasing period

    selected_disposal_rates = {
        (ob.dev_lag(), ob.metadata): ob["disposal_rate"]
        for ob in tri_w_disposal_rates.right_edge
    }

    def _get_adjusted_paid_loss(cell: Cell):
        try:
            selected_disposal_rate = selected_disposal_rates[
                (cell.dev_lag(), cell.metadata)
            ]
        except KeyError:
            selected_disposal_rate = _get_max_period_cell_for_dev_lag(cell)[
                "disposal_rate"
            ]
        cell_idx = tri_w_disposal_rates.cells.index(cell)

        if cell["disposal_rate"] == selected_disposal_rate:
            return cell["paid_loss"]
        elif cell["disposal_rate"] > selected_disposal_rate:
            # Settlement is faster than selected, so adjust losses downwards

            # Get cell in same period where dev lag is largest that's < current cell's dev_lag.
            # Defaults to a simple "cell dictionary" with 0 paid loss to handle
            # when cell is leftmost cell in a period.
            prev_cell = (
                tri_w_disposal_rates[cell_idx - 1]
                if (cell_idx > 0)
                and (tri_w_disposal_rates[cell_idx - 1].metadata == cell.metadata)
                else {"paid_loss": 0}
            )

            # The previous disposal rate is 0 if `cell` is the leftmost cell in the period
            if type(prev_cell) is dict:
                prev_selected_disposal_rate = 0
            else:
                try:
                    prev_selected_disposal_rate = selected_disposal_rates[
                        (prev_cell.dev_lag(), prev_cell.metadata)
                    ]
                except KeyError:
                    prev_selected_disposal_rate = _get_max_period_cell_for_dev_lag(
                        prev_cell
                    )["disposal_rate"]

            return np.interp(
                selected_disposal_rate,
                [prev_selected_disposal_rate, cell["disposal_rate"]],
                [prev_cell["paid_loss"], cell["paid_loss"]],
            )
        else:
            # Settlement is slower than selected, so adjust losses upwards

            # Get cell in same period where dev lag is smallest that's > current cell's dev_lag
            next_cell = (
                tri_w_disposal_rates[cell_idx + 1]
                if (cell_idx < (len(tri_w_disposal_rates) - 1))
                and (tri_w_disposal_rates[cell_idx + 1].metadata == cell.metadata)
                else {"paid_loss": 0}
            )

            try:
                next_selected_disposal_rate = selected_disposal_rates[
                    (next_cell.dev_lag(), next_cell.metadata)
                ]
            except KeyError:
                next_selected_disposal_rate = _get_max_period_cell_for_dev_lag(
                    next_cell
                )["disposal_rate"]

            return np.interp(
                selected_disposal_rate,
                [cell["disposal_rate"], next_selected_disposal_rate],
                [cell["paid_loss"], next_cell["paid_loss"]],
            )

    return tri_w_disposal_rates.derive_fields(paid_loss=_get_adjusted_paid_loss)


def reported_bs_adjustment(
    triangle: Triangle,
    annual_severity_trend: float = 0.0,
    sev_trend_method: str | None = None,
):
    """
    Adjust reported losses in `triangle` using "reported" Berquist-Sherman adjustment for changes in case
    reserve adequacy. This method detrends average case outstanding values from the latest diagonal
    using a severity trend.

    NOTE: the current implementation has the option to calculate a severity trend from the data. This
    solution is deterministic and non-ideal. Instead, severity trends should be modelled and provided
    to this method.

    Args:
        triangle: The Triangle for which reported_loss should be adjusted. Note that this triangle must
            include `reported_loss`, `paid_loss`, `open_claims`, and `cwp_claims`.
        annual_severity_trend: The annualized severity trend that will be used to adjust `reported_loss`
            provided as a decimal (e.g. 0.15 is 15% severity trend). Ignored if `sev_trend_method`
            is not None.
        sev_trend_method: Optional method used to calculate severity trend. Options are "latest" or
            "all". If "all", the mean of calculate severity trend for all cells will be used as the
            severity trend; if "latest", instead the mean will be taken only of the latest diagonal
            of `triangle`.

    Returns:
        A triangle with adjusted `reported_loss` values.
    """
    warnings.warn(
        "WARNING: Reported Berquist-Sherman is still under developement, so it may not yet work as intended."
    )

    # Adjustments are made on a calendar period-basis, so the triangle must be
    # regular and not have any experience gaps.
    if not triangle.is_regular() and len(triangle.experience_gaps):
        raise Exception("Severity trends can only be calculated on regular triangles.")

    if triangle.is_incremental:
        triangle = to_cumulative(triangle)

    tri_w_avg_case = triangle.derive_fields(
        average_case_os=lambda ob: (ob["reported_loss"] - ob["paid_loss"])
        / ob["open_claims"],
        average_paid_severity=lambda ob: ob["paid_loss"] / ob["cwp_claims"],
    )

    if sev_trend_method:
        annual_severity_trend = _sev_trend_from_hist_tri(
            tri_w_avg_case, method=sev_trend_method
        )

    def _detrend_average_case_os(cell: Cell):
        # Get the cell in the latest period with the same dev_lag and metadata as the current cell.
        # NOTE this method (in contrast to reduction) relies on triangle.cells being sorted by increasing
        # period so the last cell matching dev_lag and metadata will be the cell with the most
        # recent period.
        diagonal_cell = tri_w_avg_case.right_edge.filter(
            lambda ob: (ob.dev_lag() == cell.dev_lag())
            and (ob.metadata == cell.metadata)
        )[-1]

        trend_period = dev_lag_months(cell.period_end, diagonal_cell.period_end)

        return diagonal_cell["average_case_os"] / (1 + annual_severity_trend) ** (
            trend_period / 12
        )

    return tri_w_avg_case.derive_fields(
        average_case_os=_detrend_average_case_os,
        reported_loss=lambda ob: ob["average_case_os"] * ob["open_claims"]
        + ob["paid_loss"],
    )


def _sev_trend_from_hist_tri(
    triangle: Triangle,
    method: str = "all",  # either "all" or "latest"
):
    if method not in ["all", "latest"]:
        raise Exception(
            f"Unrecognized method `{method}` passed in severity trend calculation."
        )

    def _calculate_sev_trend(cell: Cell):
        # Get cell with the same dev lag in the nearest future period
        # NOTE this method (in contrast to reduction) relies on triangle.cells being sorted by increasing
        # period so the first cell matching dev_lag and metadata will be the cell with the period
        # immediately after the current cell's period.
        future_cells = triangle.filter(
            lambda ob: (ob.dev_lag() == cell.dev_lag()) and (ob.period > cell.period)
        )
        future_cell = (
            future_cells[0] if len(future_cells) else {"average_paid_severity": np.nan}
        )

        trend_period = (
            0
            if type(future_cell) is dict
            else dev_lag_months(cell.period_end, future_cell.period_end)
        )

        return (
            np.nan
            if np.isnan(future_cell["average_paid_severity"])
            else (future_cell["average_paid_severity"] / cell["average_paid_severity"])
            ** (trend_period / 12)
            - 1
        )

    tri_w_sev_trend = triangle.derive_fields(
        severity_trend=_calculate_sev_trend
    ).filter(lambda ob: not np.isnan(ob["severity_trend"]))

    if method == "all":
        return np.mean([ob["severity_trend"] for ob in tri_w_sev_trend])
    elif method == "latest":
        return np.mean([ob["severity_trend"] for ob in tri_w_sev_trend.right_edge])


def weight_geometric_decay(
    triangle: Triangle,
    annual_decay_factor: float,
    basis: str = "evaluation",
    tri_fields: str | list[str] | None = None,
    weight_as_field: bool = True,
) -> Triangle:
    """Implements loss experience period geometric decay weighting. Suppose a triangle's loss experience periods are indexed by t=1:T,
    with T being the latest experience period in the triangle. Then, each loss experience period is weighted by a geometric progression (r)^{(T-t)*frequency},
    where r is the geometric factor and frequency is defined as (resolution of tri) / 12.

    Note that if weight_as_field = False, the processed triangle returned from this function will have fields that are meaningless. The usage
    of this function should be limited towards the preprocessing of triangle data used when fitting loss development models if weight_as_field is False.

    Arguments:
        triangle: triangle input data.
        annual_decay_factor: the geometric factor to apply on an annual basis. Annual rates are
        automatically converted to equal the implied rates that match the resolution of tri. Note that a factor of 1
        (and technically 0) will apply equal weighting to all experience periods, i.e. the triangle will not be modified. Must
        be > 0 and <= 1.
        basis: what basis to weight the triangle on? Available options are `evaluation` (weighting is applied diagonally, i.e. by calendar year)
        and `experience` (weights are applied by column). Default is evaluation period weighting.
        tri_fields: loss fields in tri to geometrically weight. Default is None, in which case
        all fields in the triangle are weighted accordingly.
        weight_as_field: should the weights be returned as an additional field (True) or should the data be weighted in the triangle directly (False)?

    Returns:
        A triangle with a field called geometric_weight (if weight_as_field is True), or a triangle with directly weighted fields.
    """

    # Input checks and setup

    ## Annual decay factor
    if isinstance(annual_decay_factor, float):
        if annual_decay_factor > 1 or annual_decay_factor <= 0:
            raise ValueError("annual_decay_factor must be a scalar > 0 and <= 1")

    ## loss_fields
    if tri_fields is None:
        tri_fields = triangle.fields

    elif isinstance(tri_fields, str):
        if tri_fields not in triangle.fields:
            raise ValueError(f"{tri_fields} not found in tri")
        else:
            tri_fields = [tri_fields]

    elif isinstance(tri_fields, list):
        if not set(tri_fields).issubset(set(triangle.fields)):
            raise ValueError(f"At least one of {tri_fields} not found in `triangle`")

    last_date = (
        triangle.evaluation_dates[-1]
        if basis == "evaluation"
        else triangle.periods[-1][0]
    )

    if weight_as_field:
        triangle_with_weights = triangle.derive_fields(
            geometric_weight=lambda cel: _cell_weight(
                cel, last_date, annual_decay_factor, basis
            )
        )

        return triangle_with_weights
    else:
        decay_weighted_triangle = Triangle(
            [
                _reweight_cell(
                    cell,
                    last_date,
                    annual_decay_factor,
                    fields=tri_fields,
                    weighting_basis=basis,
                )
                for cell in triangle.cells
            ]
        )

        return decay_weighted_triangle


def _cell_weight(cell: Cell, last_date_start, decay_factor, weighting_basis) -> float:
    cell_date = (
        cell.evaluation_date if weighting_basis == "evaluation" else cell.period_start
    )
    months_from_end = month_to_id(last_date_start) - month_to_id(cell_date)
    years_from_end = months_from_end / 12
    return decay_factor**years_from_end


def _reweight_cell(
    cell: Cell,
    last_date_start,
    decay_factor,
    fields: list[str],
    weighting_basis,
) -> float:
    new_weight = _cell_weight(cell, last_date_start, decay_factor, weighting_basis)
    for field in fields:
        cell = cell.derive_fields(**{field: lambda ob: ob[field] * new_weight})
    return cell



import string

import altair as alt
from babel.numbers import get_currency_symbol
import pandas as pd
import numpy as np
from typing import Callable, Any, Literal

from .triangle import Triangle, Cell
from .base import metadata_diff

alt.renderers.enable("browser")

SLICE_TITLE_KWARGS = {
    "anchor": "middle",
    "font": "sans-serif",
    "fontWeight": "normal",
    "fontSize": 12,
}

BASE_HEIGHT = 600
BASE_WIDTH = "container"

BASE_AXIS_LABEL_FONT_SIZE = 16
BASE_AXIS_TITLE_FONT_SIZE = 18
FONT_SIZE_DECAY_FACTOR = 0.2

CellArgs = Cell | Cell, Cell, Cell
MetricFunc = Callable[[CellArgs], float | int | np.ndarray]
MetricFuncDict = dict[str, MetricFunc]

COMMON_METRIC_DICT: MetricFuncDict = {
    "Paid Loss Ratio": lambda cell: 100 * cell["paid_loss"] / cell["earned_premium"],
    "Reported Loss Ratio": lambda cell: 100 * cell["reported_loss"] / cell["earned_premium"],
    "Incurred Loss Ratio": lambda cell: 100 * cell["incurred_loss"] / cell["earned_premium"],
    "Paid Loss": lambda cell: cell["paid_loss"] / 1e6,
    "Reported Loss": lambda cell: cell["reported_loss"] / 1e6,
    "Incurred Loss": lambda cell: cell["incurred_loss"] / 1e6,
    "Paid ATA": lambda cell, prev_cell: cell["paid_loss"] / prev_cell["paid_loss"],
    "Reported ATA": lambda cell, prev_cell: cell["reported_loss"] / prev_cell["reported_loss"],
    "Paid Incremental ATA": lambda cell, prev_cell: cell["paid_loss"] / prev_cell["paid_loss"] - 1,
    "Reported Incremental ATA": lambda cell, prev_cell: cell["reported_loss"] / prev_cell["reported_loss"] - 1,
}

MetricFuncSpec = MetricFuncDict | str | list[str]


def _resolve_metric_spec(metric_spec: MetricFuncSpec) -> MetricFuncDict:
    if isinstance(metric_spec, str):
        metric_spec = [metric_spec]
    if isinstance(metric_spec, list):
        result = {}
        for ref in metric_spec:
            if not isinstance(ref, str):
                raise ValueError("Supplied metric references must be strings")
            elif ref not in COMMON_METRIC_DICT:
                raise ValueError(f"Don't know the definition of metric {ref}")
            else:
                result[ref] = COMMON_METRIC_DICT[ref]
        return result
    else:
        return metric_spec


@alt.theme.register("bermuda_plot_theme", enable=True)
def bermuda_plot_theme() -> alt.theme.ThemeConfig:
    return {
        "autosize": {"contains": "content", "resize": True},
        "config": {
            "style": {
                "group-title": {"fontSize": 24},
                "group-subtitle": {"fontSize": 18},
                "guide-label": {
                    "fontSize": BASE_AXIS_LABEL_FONT_SIZE,
                    "font": "sans-serif",
                },
                "guide-title": {
                    "fontSize": BASE_AXIS_TITLE_FONT_SIZE,
                    "font": "sans-serif",
                },
            },
            "mark": {"color": "black"},
            "title": {"anchor": "start", "offset": 20},
            "legend": {
                "orient": "right",
                "titleAnchor": "start",
                "layout": {
                    "direction": "vertical",
                },
            },
        },
    }


def plot_right_edge(
    triangle: Triangle,
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    main_title = alt.Title(
        "Latest Loss Ratio", subtitle="The most recent loss ratio diagonal"
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        triangle,
        plot_func=_plot_right_edge,
        title=main_title,
        facet_titles=facet_titles,
        uncertainty=uncertainty,
        uncertainty_type=uncertainty_type,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig


def _plot_right_edge(
    triangle: Triangle,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
) -> alt.Chart:
    if "earned_premium" not in triangle.fields:
        raise ValueError(
            "Triangle must contain `earned_premium` to plot its right edge. "
            f"This triangle contains {triangle.fields}"
        )

    loss_fields = [field for field in triangle.fields if "_loss" in field]

    loss_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    **_calculate_field_summary(
                        cell=cell,
                        prev_cell=None,
                        func=lambda ob: ob[field] / ob["earned_premium"],
                        name="loss_ratio",
                    ),
                    "Field": field.replace("_loss", "").title() + " LR",
                }
                for cell in triangle.right_edge
                for field in loss_fields
                if "earned_premium" in cell
            ]
        ]
    )

    premium_data = alt.Data(
        values=[
            *[
                {
                    "period_start": pd.to_datetime(cell.period_start),
                    "period_end": pd.to_datetime(cell.period_end),
                    "evaluation_date": pd.to_datetime(cell.evaluation_date),
                    "dev_lag": cell.dev_lag(),
                    "Earned Premium": cell["earned_premium"],
                    "Field": "Earned Premium",
                }
                for cell in triangle.right_edge
                if "earned_premium" in cell
            ]
        ]
    )

    currency = _currency_symbol(triangle)

    bar = (
        alt.Chart(premium_data, title=title)
        .mark_bar()
        .encode(
            x=alt.X("yearmonth(period_start):O"),
            y=alt.Y("Earned Premium:Q").axis(format="$.2s"),
            color=alt.Color("Field:N").scale(range=["lightgray"]),
            tooltip=[
                alt.Tooltip("period_start:T", title="Period Start"),
                alt.Tooltip("period_end:T", title="Period End"),
                alt.Tooltip("dev_lag:O", title="Dev Lag"),
                alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                alt.Tooltip("Earned Premium:Q", format=f"{currency},.0f"),
            ],
        )
    )

    if uncertainty and uncertainty_type == "ribbon":
        loss_error = (
            alt.Chart(loss_data)
            .mark_area(
                opacity=0.5,
            )
            .encode(
                x=alt.X("yearmonth(period_start):T"),
                y=alt.Y("loss_ratio_lower_ci:Q").axis(title="Loss Ratio %", format="%"),
                y2=alt.Y2("loss_ratio_upper_ci:Q"),
                color=alt.Color("Field:N"),
            )
        )
    elif uncertainty and uncertainty_type == "segments":
        loss_error = (
            alt.Chart(loss_data)
            .mark_errorbar(thickness=3)
            .encode(
                x=alt.X("yearmonth(period_start):T").title("Period Start"),
                y=alt.Y("loss_ratio_lower_ci:Q").axis(title="Loss Ratio %", format="%"),
                y2=alt.Y2("loss_ratio_upper_ci:Q"),
                color=alt.Color("Field:N"),
            )
        )
    else:
        loss_error = alt.LayerChart()

    lines = (
        alt.Chart(loss_data)
        .mark_line(
            size=1,
        )
        .encode(
            x=alt.X("yearmonth(period_start):T", axis=alt.Axis(labelAngle=0)).title(
                "Period Start"
            ),
            y=alt.Y(
                "loss_ratio:Q", scale=alt.Scale(zero=True), axis=alt.Axis(format="%")
            ).title("Loss Ratio %"),
            color=alt.Color("Field:N"),
        )
    )

    points = (
        alt.Chart(loss_data)
        .mark_point(
            size=max(20, 100 / mark_scaler),
            filled=True,
            opacity=1,
        )
        .encode(
            x=alt.X("yearmonth(period_start):T", axis=alt.Axis(labelAngle=0)).title(
                "Period Start"
            ),
            y=alt.Y(
                "loss_ratio:Q", scale=alt.Scale(zero=True), axis=alt.Axis(format="%")
            ),
            color=alt.Color("Field:N").legend(title=None),
            tooltip=[
                alt.Tooltip("period_start:T", title="Period Start"),
                alt.Tooltip("period_end:T", title="Period End"),
                alt.Tooltip("dev_lag:O", title="Dev Lag"),
                alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                alt.Tooltip("loss_ratio:Q", title="Loss Ratio (%)", format=".2%"),
            ],
        )
    )

    fig = alt.layer(bar, loss_error + lines + points).resolve_scale(
        y="independent",
        color="independent",
    )

    return fig.interactive()


def plot_data_completeness(
    triangle: Triangle,
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    main_title = alt.Title(
        "Triangle Completeness",
        subtitle="The number of data fields available per cell",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        triangle,
        plot_func=_plot_data_completeness,
        title=main_title,
        facet_titles=facet_titles,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig


def _plot_data_completeness(
    triangle: Triangle, title: alt.Title, mark_scaler: int
) -> alt.Chart:
    if not triangle.is_disjoint:
        raise Exception(
            "This triangle isn't disjoint! You probably don't want to use it"
        )
    if not triangle.is_semi_regular:
        raise Exception(
            "This triangle isn't semi-regular! You probably don't want to use it"
        )

    currency = _currency_symbol(triangle)

    selection = alt.selection_point()

    cell_data = alt.Data(
        values=[
            *[
                {
                    "period_start": pd.to_datetime(cell.period_start),
                    "period_end": pd.to_datetime(cell.period_end),
                    "evaluation_date": pd.to_datetime(cell.evaluation_date),
                    "dev_lag": cell.dev_lag(),
                    "Number of Fields": len(cell.values),
                    "Fields": ", ".join(
                        [
                            field.replace("_", " ").title()
                            + f" ({currency}{np.mean(cell[field]):,.0f})"
                            for field in cell.values
                        ]
                    ),
                }
                for cell in triangle.cells
            ]
        ]
    )

    fig = (
        alt.Chart(
            cell_data,
            title=title,
        )
        .mark_circle(size=500 * 1 / mark_scaler, opacity=1)
        .encode(
            alt.X(
                "dev_lag:N", axis=alt.Axis(labelAngle=0), scale=alt.Scale(zero=True)
            ).title("Dev Lag (months)"),
            alt.Y(
                "yearmonth(period_start):T", scale=alt.Scale(padding=15, reverse=True)
            ).title("Period Start"),
            color=alt.condition(
                selection,
                alt.Color("Number of Fields:N").scale(scheme="dark2"),
                alt.value("lightgray"),
            ),
            tooltip=[
                alt.Tooltip("period_start:T", title="Period Start"),
                alt.Tooltip("period_end:T", title="Period End"),
                alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                alt.Tooltip("dev_lag:N", title="Dev Lag (months)"),
                alt.Tooltip("Fields:N"),
            ],
        )
        .add_params(selection)
    )

    return fig.interactive()


def plot_heatmap(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Loss Ratio"],
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a heatmap."""
    main_title = alt.Title(
        "Triangle Heatmap",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            triangle=triangle,
            plot_func=_plot_heatmap,
            metric_dict=metric_dict,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            mark_scaler=max_cols,
            ncols=max_cols,
        )
        .resolve_scale(color="independent")
        .resolve_legend(color="independent")
    )
    return fig


def _plot_heatmap(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    title: alt.Title,
    mark_scaler: int,
) -> alt.Chart:
    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    **_calculate_field_summary(cell, prev_cell, metric, "metric"),
                    "Field": name,
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X("dev_lag:N", axis=alt.Axis(labelAngle=0)).title("Dev Lag (months)"),
        y=alt.X("yearmonth(period_start):O", scale=alt.Scale(reverse=False)).title(
            "Period Start"
        ),
    )

    stroke_predicate = alt.datum.metric_sd / alt.datum.metric > 0
    selection = alt.selection_interval()
    heatmap = (
        base.mark_rect()
        .encode(
            color=alt.when(selection)
            .then(
                alt.Color("metric:Q", scale=alt.Scale(range=MANAGUA_VALS)).title(name)
            )
            .otherwise(
                alt.value("gray"),
            ),
            tooltip=[
                alt.Tooltip("period_start:T", title="Period Start"),
                alt.Tooltip("period_end:T", title="Period End"),
                alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                alt.Tooltip("dev_lag:O", title="Dev Lag (months)"),
                alt.Tooltip("metric:Q", title=name),
            ],
            stroke=alt.when(stroke_predicate).then(alt.value("black")),
            strokeWidth=alt.when(stroke_predicate)
            .then(alt.value(3))
            .otherwise(alt.value(0)),
        )
        .add_params(selection)
    )

    text = base.mark_text(
        fontSize=BASE_AXIS_TITLE_FONT_SIZE
        * np.exp(-FONT_SIZE_DECAY_FACTOR * mark_scaler),
        font="sans-serif",
    ).encode(
        text=alt.Text("metric:Q", format=",.1f")
    )

    return (heatmap + text).resolve_scale(color="independent")


def plot_atas(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid ATA"],
    ncols: int | None = None,
    width: int = 400,
    height: int = 200,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle ATAs."""
    main_title = alt.Title(
        "Triangle ATAs",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            triangle,
            plot_func=_plot_atas,
            metric_dict=metric_dict,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols))
        .resolve_scale(color="independent")
    )
    return fig


def _plot_atas(
    triangle: Triangle, metric: MetricFunc, name: str, title: alt.Title
) -> alt.Chart:
    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    **_calculate_field_summary(cell, prev_cell, metric, "metric"),
                    "Field": name,
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    tooltip = [
        alt.Tooltip("period_start:T", title="Period Start"),
        alt.Tooltip("period_end:T", title="Period End"),
        alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
        alt.Tooltip("dev_lag:O", title="Dev Lag (months)"),
        alt.Tooltip("metric:Q", title=name, format=".2f"),
    ]

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X("dev_lag:N", axis=alt.Axis(labelAngle=0))
        .title("Dev Lag (months)")
        .scale(padding=10),
        y=alt.X("metric:Q").title(name).scale(zero=False, padding=10),
        tooltip=tooltip,
    )

    points = base.mark_point(color="black", filled=True)
    boxplot = base.mark_boxplot(
        opacity=0.7,
        color="skyblue",
        median=alt.MarkConfig(stroke="black"),
        rule=alt.MarkConfig(stroke="black"),
        box=alt.MarkConfig(stroke="black"),
    )

    return (points + boxplot).interactive()


def plot_growth_curve(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Loss Ratio"],
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a growth curve."""
    main_title = alt.Title(
        "Triangle Growth Curve",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            triangle,
            plot_func=_plot_growth_curve,
            metric_dict=metric_dict,
            uncertainty=uncertainty,
            uncertainty_type=uncertainty_type,
            mark_scaler=max_cols,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols))
    )
    return fig


def _plot_growth_curve(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    uncertainty_type: Literal["ribbon", "segments"],
) -> alt.Chart:
    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    "last_lag": max(
                        triangle.filter(lambda ob: ob.period == cell.period).dev_lags()
                    ),
                    **_calculate_field_summary(cell, prev_cell, metric, "metric"),
                    "Field": name,
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    color = (
        alt.Color("yearmonth(period_start):Q")
        .scale(range=MANAGUA_VALS)
        .legend(title="Period Start")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X("dev_lag:Q", axis=alt.Axis(grid=True, labelAngle=0))
        .title("Dev Lag (months)"),
        y=alt.X("metric:Q").title(name),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip("metric:Q", format=",.1f", title=name),
        ],
    )

    lines = base.mark_line(opacity=0.2).encode(
        color=color_conditional_no_legend
    )
    points = base.mark_point(stroke=None, filled=True).encode(
        color=color_conditional_no_legend,
        opacity=opacity_conditional,
    )
    ultimates = (
        base.mark_point(size=300 / mark_scaler, filled=True, stroke=None)
        .encode(
            color=color_conditional,
            opacity=opacity_conditional,
            strokeOpacity=opacity_conditional,
        )
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty and uncertainty_type == "ribbon":
        ribbon_opacity_conditional = (
            alt.when(selector)
            .then(alt.OpacityValue(0.5))
            .otherwise(alt.OpacityValue(0.2))
        )
        errors = base.mark_area(
            opacity=0.5,
        ).encode(
            y=alt.Y("metric_lower_ci:Q"),
            y2=alt.Y2("metric_upper_ci:Q"),
            color=color_conditional_no_legend,
            opacity=ribbon_opacity_conditional,
        )
    elif uncertainty and uncertainty_type == "segments":
        errors = base.mark_errorbar(thickness=5).encode(
            y=alt.Y("metric_lower_ci:Q").axis(title=name),
            y2=alt.Y2("metric_upper_ci:Q"),
            color=color_conditional_no_legend,
            opacity=opacity_conditional,
        )
    else:
        errors = alt.LayerChart()

    return (
        alt.layer(errors + lines + points, ultimates.add_params(selector))
        .resolve_scale(color="independent")
        .interactive()
    )


def plot_sunset(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Incremental ATA"],
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a sunset."""
    main_title = alt.Title(
        "Triangle Sunset",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            triangle,
            plot_func=_plot_sunset,
            metric_dict=metric_dict,
            uncertainty=uncertainty,
            uncertainty_type=uncertainty_type,
            mark_scaler=max_cols,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols))
    )
    return fig.interactive()


def _plot_sunset(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    uncertainty_type: Literal["ribbon", "segments"],
) -> alt.Chart:
    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    **_calculate_field_summary(cell, prev_cell, metric, name),
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    color = (
        alt.Color("dev_lag:Q")
        .scale(range=MANAGUA_VALS)
        .legend(title="Development Lag")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["dev_lag"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X(
            "yearmonth(evaluation_date):O", axis=alt.Axis(grid=True, labelAngle=0)
        ).title("Calendar Year"),
        y=alt.X(f"{name}:Q").title(name).scale(type="pow", exponent=0.3),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{name}:Q", format=",.1f", title=name),
        ],
    )

    points = base.mark_point(
        stroke=None, size=30 / mark_scaler, filled=True
    ).encode(
        color=color_conditional,
        opacity=opacity_conditional,
        strokeOpacity=opacity_conditional,
    )
    regression = (
        base.transform_loess(
            "evaluation_date", f"{name}", groupby=["dev_lag"], bandwidth=0.6
        )
        .mark_line(strokeWidth=2)
        .encode(color=color_conditional_no_legend, opacity=opacity_conditional)
    )

    if uncertainty and uncertainty_type == "ribbon":
        ribbon_conditional = (
            alt.when(selector)
            .then(alt.OpacityValue(0.5))
            .otherwise(alt.OpacityValue(0.2))
        )
        errors = base.mark_area().encode(
            y=alt.Y(f"{name}_lower_ci:Q").axis(title=name),
            y2=alt.Y2(f"{name}_upper_ci:Q"),
            color=color_conditional_no_legend,
            opacity=ribbon_conditional,
        )
    elif uncertainty and uncertainty_type == "segments":
        errors = base.mark_errorbar(thickness=5).encode(
            y=alt.Y(f"{name}_lower_ci:Q").axis(title=name),
            y2=alt.Y2(f"{name}_upper_ci:Q"),
            color=color_conditional_no_legend,
            opacity=opacity_conditional,
        )
    else:
        errors = alt.LayerChart()

    return (
        alt.layer(errors, regression, points)
        .add_params(selector)
        .resolve_scale(color="independent")
    )


def plot_mountain(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Loss Ratio"],
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a mountain."""
    main_title = alt.Title(
        "Triangle Mountain Plot",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            triangle,
            plot_func=_plot_mountain,
            metric_dict=metric_dict,
            uncertainty=uncertainty,
            uncertainty_type=uncertainty_type,
            mark_scaler=max_cols,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols))
    )
    return fig.interactive()


def _plot_mountain(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    uncertainty_type: Literal["ribbon", "segments"],
) -> alt.Chart:
    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    "last_lag": max(
                        triangle.filter(lambda ob: ob.period == cell.period).dev_lags()
                    ),
                    **_calculate_field_summary(cell, prev_cell, metric, "metric"),
                    "Field": name,
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    color = (
        alt.Color("dev_lag:Q")
        .scale(range=MANAGUA_VALS)
        .legend(title="Development Lag (months)")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["dev_lag"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X(
            "yearmonth(period_start):O", axis=alt.Axis(grid=True, labelAngle=0)
        ).title("Period Start"),
        y=alt.X("metric:Q").title(name),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip("metric:Q", format=",.1f", title=name),
        ],
    )

    lines = base.mark_line().encode(
        color=color_conditional_no_legend, opacity=opacity_conditional
    )
    points = base.mark_point(filled=True, stroke=None).encode(
        color=color_conditional,
        opacity=opacity_conditional,
    )

    if uncertainty and uncertainty_type == "ribbon":
        ribbon_conditional = (
            alt.when(selector)
            .then(alt.OpacityValue(0.5))
            .otherwise(alt.OpacityValue(0.2))
        )
        errors = base.mark_area().encode(
            y=alt.Y("metric_lower_ci:Q"),
            y2=alt.Y2("metric_upper_ci:Q"),
            color=color_conditional_no_legend,
            opacity=ribbon_conditional,
        )
    elif uncertainty and uncertainty_type == "segments":
        errors = base.mark_errorbar(thickness=5).encode(
            y=alt.Y("metric_lower_ci:Q").axis(title=name),
            y2=alt.Y2("metric_upper_ci:Q"),
            color=color_conditional_no_legend,
            opacity=opacity_conditional,
        )
    else:
        errors = alt.LayerChart()

    return alt.layer(errors + lines, points.add_params(selector)).resolve_scale(
        color="independent"
    )


def plot_ballistic(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Paid Loss Ratio": lambda cell: 100
        * cell["paid_loss"]
        / cell["earned_premium"],
        "Reported Loss Ratio": lambda cell: 100
        * cell["reported_loss"]
        / cell["earned_premium"],
    },
    uncertainty: bool = True,
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a ballistic."""
    main_title = alt.Title(
        "Triangle Ballistic Plot",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        triangle,
        plot_func=_plot_ballistic,
        axis_metrics=axis_metrics,
        title=main_title,
        facet_titles=facet_titles,
        uncertainty=uncertainty,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig


def _plot_ballistic(
    triangle: Triangle,
    axis_metrics: MetricFuncDict,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
) -> alt.Chart:
    (name_x, name_y), (func_x, func_y) = zip(*axis_metrics.items())

    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    "last_lag": max(
                        triangle.filter(lambda ob: ob.period == cell.period).dev_lags()
                    ),
                    **_calculate_field_summary(cell, prev_cell, func_x, name_x),
                    **_calculate_field_summary(cell, prev_cell, func_y, name_y),
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    color = (
        alt.Color("dev_lag:Q")
        .scale(range=MANAGUA_VALS)
        .legend(title="Development Lag (months)")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X(f"{name_x}:Q").title(name_x).axis(grid=True),
        y=alt.X(f"{name_y}:Q").title(name_y).axis(grid=True),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:O", title="Dev Lag (months)"),
            alt.Tooltip(f"{name_x}:Q", format=".1f"),
            alt.Tooltip(f"{name_y}:Q", format=".1f"),
        ],
    )

    diagonal = (
        alt.Chart(metric_data)
        .mark_line(color="black", strokeDash=[5, 5])
        .encode(
            x=f"{name_x}:Q",
            y=f"{name_x}:Q",
        )
    )

    lines = base.mark_line(color="black", strokeWidth=0.5).encode(
        detail="period_start:N", opacity=opacity_conditional
    )
    points = base.mark_point(
        filled=True, size=100 / mark_scaler, stroke=None
    ).encode(color=color_conditional, opacity=opacity_conditional)
    ultimates = (
        base.mark_point(size=200 / mark_scaler, filled=True, stroke=None)
        .encode(color=color_conditional, opacity=opacity_conditional)
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty:
        errors = base.mark_errorbar(thickness=5).encode(
            y=alt.Y(f"{name_y}_lower_ci:Q").axis(title=name_y),
            y2=alt.Y2(f"{name_y}_upper_ci:Q"),
            color=color_conditional_no_legend,
        )
    else:
        errors = alt.LayerChart()

    return (
        alt.layer(diagonal, errors + lines, (points + ultimates).add_params(selector))
        .resolve_scale(color="independent")
        .interactive()
    )


def plot_broom(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Paid/Reported Ratio": lambda cell: cell["paid_loss"] / cell["reported_loss"],
        "Paid Loss Ratio": lambda cell: 100
        * cell["paid_loss"]
        / cell["earned_premium"],
    },
    rule: int | None = 1,
    uncertainty: bool = True,
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a broom."""
    main_title = alt.Title(
        "Triangle Broom Plot",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        triangle,
        plot_func=_plot_broom,
        axis_metrics=axis_metrics,
        title=main_title,
        facet_titles=facet_titles,
        uncertainty=uncertainty,
        rule=rule,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig.interactive()


def _plot_broom(
    triangle: Triangle,
    axis_metrics: MetricFuncDict,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    rule: int | None,
) -> alt.Chart:
    (name_x, name_y), (func_x, func_y) = zip(*axis_metrics.items())

    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    "last_lag": max(
                        triangle.filter(lambda ob: ob.period == cell.period).dev_lags()
                    ),
                    **_calculate_field_summary(cell, prev_cell, func_x, name_x),
                    **_calculate_field_summary(cell, prev_cell, func_y, name_y),
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    color = (
        alt.Color("dev_lag:Q")
        .scale(range=MANAGUA_VALS)
        .legend(title="Development Lag (months)")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X(f"{name_x}:Q").scale(padding=10, nice=False).title(name_x),
        y=alt.Y(f"{name_y}:Q").title(name_y),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{name_x}:Q", format=".1f"),
            alt.Tooltip(f"{name_y}:Q", format=".1f"),
        ],
    )

    wall = (
        alt.Chart().mark_rule(strokeDash=[12, 5], opacity=0.5, strokeWidth=2)
    ).encode()
    if rule is not None:
        wall = wall.encode(x=alt.datum(rule))

    lines = base.mark_line(color="black", strokeWidth=0.5).encode(
        detail="period_start:N", opacity=opacity_conditional
    )
    points = base.mark_point(
        filled=True, size=100 / mark_scaler, stroke=None
    ).encode(
        color=color_conditional,
        opacity=opacity_conditional,
        strokeOpacity=opacity_conditional,
    )
    ultimates = (
        base.mark_point(size=300 / mark_scaler, filled=True, stroke=None)
        .encode(
            color=color_conditional,
            opacity=opacity_conditional,
            strokeOpacity=opacity_conditional,
        )
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty:
        errors = base.mark_errorbar(thickness=5).encode(
            x=alt.X(f"{name_x}_lower_ci:Q").axis(title=name_x),
            x2=alt.X2(f"{name_x}_upper_ci:Q"),
            color=color_conditional_no_legend,
        )
    else:
        errors = alt.LayerChart()

    return alt.layer(
        errors + lines + wall, (points + ultimates).add_params(selector)
    ).resolve_scale(color="independent")


def plot_drip(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Reported Loss Ratio": lambda cell: 100
        * cell["reported_loss"]
        / cell["earned_premium"],
        "Open Claim Share": lambda cell: 100
        * cell["open_claims"]
        / cell["reported_claims"],
    },
    uncertainty: bool = True,
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a drip."""
    main_title = alt.Title(
        "Triangle Drip Plot",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = (
        _build_metric_slice_charts(
            triangle,
            plot_func=_plot_drip,
            axis_metrics=axis_metrics,
            title=main_title,
            facet_titles=facet_titles,
            uncertainty=uncertainty,
            width=width,
            height=height,
            mark_scaler=max_cols,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .resolve_scale(color="independent")
    )
    return fig.interactive()


def _plot_drip(
    triangle: Triangle,
    axis_metrics: MetricFuncDict,
    title: alt.Title,
    uncertainty: bool,
    mark_scaler: int,
) -> alt.Chart:
    (name_x, name_y), (func_x, func_y) = zip(*axis_metrics.items())

    metric_data = alt.Data(
        values=[
            *[
                {
                    **_core_plot_data(cell),
                    "last_lag": max(
                        triangle.filter(lambda ob: ob.period == cell.period).dev_lags()
                    ),
                    **_calculate_field_summary(cell, prev_cell, func_x, name_x),
                    **_calculate_field_summary(cell, prev_cell, func_y, name_y),
                }
                for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
            ]
        ]
    )

    color = (
        alt.Color("dev_lag:Q")
        .scale(range=MANAGUA_VALS)
        .legend(title="Development Lag (months)")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(metric_data, title=title).encode(
        x=alt.X(f"{name_x}:Q").title(name_x, padding=10),
        y=alt.Y(f"{name_y}:Q").title(name_y).scale(nice=False, padding=10),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{name_x}:Q", format=".1f"),
            alt.Tooltip(f"{name_y}:Q", format=".1f"),
        ],
    )

    lines = base.mark_line(color="black", strokeWidth=0.5).encode(
        detail="period_start:N", opacity=opacity_conditional
    )
    points = base.mark_point(
        filled=True, size=100 / mark_scaler, stroke=None
    ).encode(color=color_conditional, opacity=opacity_conditional)
    ultimates = (
        base.mark_point(size=300 / mark_scaler, filled=True, stroke=None)
        .encode(color=color_conditional, opacity=opacity_conditional)
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty:
        errors = base.mark_errorbar(thickness=5).encode(
            y=alt.Y(f"{name_y}_lower_ci:Q").title(name_y),
            y2=alt.Y2(f"{name_y}_upper_ci:Q"),
            color=color_conditional_no_legend,
        )
    else:
        errors = alt.LayerChart()

    return alt.layer(
        errors + lines, (points + ultimates).add_params(selector)
    ).resolve_scale(color="independent")


def plot_hose(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Paid Loss Ratio": lambda cell: 100
        * cell["paid_loss"]
        / cell["earned_premium"],
        "Incremental Paid Loss Ratio": lambda cell, prev_cell: 100
        * (
            cell["paid_loss"] / cell["earned_premium"]
            - prev_cell["paid_loss"] / prev_cell["earned_premium"]
        ),
    },
    uncertainty: bool = True,
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    return plot_drip(
        triangle, axis_metrics, uncertainty, width, height, ncols, facet_titles
    ).properties(title="Triangle Hose Plot")


def _build_metric_slice_charts(
    triangle, plot_func, title, facet_titles, width, height, ncols, **plot_kwargs
):
    charts = []
    n_slices = len(triangle.slices)
    for i, (_, triangle_slice) in enumerate(triangle.slices.items()):
        if facet_titles is None:
            slice_title = _slice_label(triangle_slice, triangle)
        else:
            slice_title = facet_titles[i]
        if plot_kwargs.get("metric_dict") is not None:
            for name, metric in plot_kwargs["metric_dict"].items():
                metric_title = (
                    (n_slices > 1) * (slice_title + ": ") + name
                    if facet_titles is None
                    else slice_title
                )
                charts.append(
                    plot_func(
                        triangle=triangle_slice,
                        metric=metric,
                        name=name,
                        title=alt.Title(metric_title, **SLICE_TITLE_KWARGS),
                        **{k: v for k, v in plot_kwargs.items() if k != "metric_dict"},
                    ).properties(width=width, height=height)
                )
        else:
            charts.append(
                plot_func(
                    triangle=triangle_slice,
                    title=alt.Title(slice_title, **SLICE_TITLE_KWARGS),
                    **plot_kwargs,
                ).properties(width=width, height=height)
            )
    fig = (
        _concat_charts(charts, title=title, ncols=ncols)
        .configure_axis(**_compute_font_sizes(ncols))
        .configure_legend(**_compute_font_sizes(ncols))
    )
    return fig


def _core_plot_data(cell: Cell) -> dict[str, Any]:
    return {
        "period_start": pd.to_datetime(cell.period_start),
        "period_end": pd.to_datetime(cell.period_end),
        "evaluation_date": pd.to_datetime(cell.evaluation_date),
        "dev_lag": cell.dev_lag(),
    }


def _calculate_field_summary(
    cell: Cell,
    prev_cell: Cell | None,
    func: MetricFunc,
    name: str,
    probs: tuple[float, float] = (0.05, 0.95),
):
    none_dict = {
        f"{name}": None,
        f"{name}_sd": None,
        f"{name}_lower_ci": None,
        f"{name}_upper_ci": None,
    }
    try:
        metric = func(cell, prev_cell)
        if prev_cell.period != cell.period:
            raise IndexError
    except TypeError as e:
        if "takes 1 positional argument but 2 were given" in e.args[0]:
            try:
                metric = func(cell)
            except Exception:
                return none_dict
        elif "'NoneType' object is not subscriptable" in e.args[0]:
            return none_dict
    except Exception:
        return none_dict

    if np.isscalar(metric) or len(metric) == 1:
        return {
            f"{name}": metric,
            f"{name}_sd": 0,
            f"{name}_lower_ci": None,
            f"{name}_upper_ci": None,
        }

    point = np.mean(metric)
    lower, upper = np.quantile(metric, probs)
    return {
        f"{name}": point,
        f"{name}_sd": metric.std(),
        f"{name}_lower_ci": lower,
        f"{name}_upper_ci": upper,
    }


def _compute_font_sizes(mark_scaler: int) -> dict[str, float | int]:
    return {
        "titleFontSize": BASE_AXIS_TITLE_FONT_SIZE
        * np.exp(-FONT_SIZE_DECAY_FACTOR * (mark_scaler - 1)),
        "labelFontSize": BASE_AXIS_LABEL_FONT_SIZE
        * np.exp(-FONT_SIZE_DECAY_FACTOR * (mark_scaler - 1)),
    }


def _currency_symbol(triangle: Triangle) -> str:
    code = triangle.metadata[0].currency
    return get_currency_symbol(code, locale="en_US") or "$"


def _concat_charts(charts: list[alt.Chart], ncols: int, **kwargs) -> alt.Chart:
    if len(charts) == 1:
        return charts[0].properties(**kwargs)

    fig = alt.concat(*charts, columns=ncols, **kwargs)
    return fig


def _determine_facet_cols(n: int):
    """This is a replication of grDevices::n2mfrow in R"""
    return int(min(n, np.ceil(n / np.sqrt(n))))


_MANAGUA_RAW = [
    [1, 0.81263, 0.40424],
    [0.92706, 0.69585, 0.36536],
    [0.85597, 0.58963, 0.32987],
    [0.78516, 0.4928, 0.29757],
    [0.71247, 0.40508, 0.26891],
    [0.633, 0.32425, 0.24442],
    [0.54243, 0.24891, 0.22689],
    [0.44962, 0.18873, 0.22534],
    [0.37143, 0.15932, 0.25403],
    [0.32066, 0.17283, 0.32465],
    [0.29878, 0.22656, 0.42904],
    [0.3009, 0.30548, 0.54191],
    [0.32092, 0.39385, 0.64188],
    [0.35064, 0.48323, 0.72305],
    [0.38579, 0.57766, 0.79429],
    [0.4241, 0.67996, 0.86256],
    [0.46377, 0.78936, 0.93036],
    [0.50411, 0.90708, 0.99978]
]


def _float_to_rgb(color_float: tuple[float, float, float]) -> str:
    rf, gf, bf = color_float
    r = round(rf * 255)
    g = round(gf * 255)
    b = round(bf * 255)
    return f"#{r:02x}{g:02x}{b:02x}"


MANAGUA_VALS = [_float_to_rgb(val) for val in _MANAGUA_RAW]


def _slice_label(slice_tri: Triangle, base_tri: Triangle):
    slice_metadata = metadata_diff(base_tri.common_metadata, slice_tri.common_metadata)

    # Custom elements
    custom_elems = []
    for label, value in {**slice_metadata.details, **slice_metadata.loss_details}.items():
        custom_elems.append(f"{string.capwords(label)}: {value}")

    # Bare elements
    bare_elems = []
    if slice_metadata.country is not None:
        bare_elems.append(slice_metadata.country)
    if slice_metadata.reinsurance_basis is not None:
        bare_elems.append(slice_metadata.reinsurance_basis)
    if slice_metadata.loss_definition is not None:
        bare_elems.append(slice_metadata.loss_definition)

    # Decorated elements
    decorated_elems = []
    if slice_metadata.per_occurrence_limit is not None:
        decorated_elems.append(f"limit {slice_metadata.per_occurrence_limit}")
    if slice_metadata.risk_basis is not None:
        decorated_elems.append(f"{slice_metadata.risk_basis} Basis")
    if slice_metadata.currency is not None:
        decorated_elems.append(f"in {slice_metadata.currency}")

    custom_label = ", ".join(custom_elems)
    bare_label = " ".join(bare_elems)
    decorated_label = "(" + ", ".join(decorated_elems) + ")"

    label = ""
    if custom_label:
        label += custom_label
    if label and bare_label:
        label += "; "
    if bare_label:
        label += bare_label
    if label and len(decorated_label) > 2:
        label += " "
    if len(decorated_label) > 2:
        label += decorated_label

    return label



from ..date_utils import add_months
from ..triangle import Triangle


def fill_forward_gaps(
    triangle: Triangle,
    eval_resolution: int | None = None,
    fill_with_none: bool = False,
) -> Triangle:
    """For each slice period, fill forward cells for missing evaluations
    at the current triangle resolution.

    Args:
        triangle: Triangle to fill
        eval_resolution: Evaluation resolution in months, if not supplied
            this will be calculated from the existing triangle
        fill_with_none: Fill missing cell values with None, otherwise
            copy the previous cell values
    """
    if eval_resolution is None:
        eval_resolution = triangle.eval_date_resolution
    filled_cells = []
    for _, period in triangle.period_rows:
        period_cells = {cell.dev_lag(): cell for cell in period}
        required_lags = {
            lag
            for lag in range(
                int(period[0].dev_lag()),
                int(period[-1].dev_lag() + eval_resolution),
                eval_resolution,
            )
        }
        new_lags = required_lags - set(period_cells.keys())
        for lag in sorted(new_lags):
            period_cells[lag] = period_cells[lag - eval_resolution].replace(
                evaluation_date=lambda cell: add_months(cell.period_end, lag),
            )
            if fill_with_none:
                period_cells[lag] = period_cells[lag].replace(
                    values={k: None for k in period_cells[lag].values}
                )
        filled_cells.extend(period_cells.values())

    return Triangle(filled_cells)



from datetime import date
from typing import Tuple, Union

import numpy as np
from dateutil.relativedelta import relativedelta
from scipy.interpolate import interp1d

from ..base import Cell, CellValue
from ..date_utils import add_months, eval_date_resolution, period_resolution
from ..errors import TriangleError
from ..triangle import Triangle
from .basis import to_cumulative, to_incremental

__all__ = [
    "disaggregate",
    "disaggregate_development",
    "disaggregate_experience",
]


DEFAULT_INTERPOLATION_FIELDS = [
    "paid_loss",
    "reported_loss",
    "incurred_loss",
    "earned_premium",
]


def disaggregate(
    triangle: Triangle,
    resolution_exp_months: int = 3,
    resolution_dev_months: int = 3,
    fields: list[str] | None = None,
    period_weights: Union[list[float], dict[date, list[float]]] | None = None,
    interpolation_method: str = "linear",
    extrapolate_first_period: bool = True,
    **interpolation_kwargs,
) -> Triangle:
    """
    Disaggregate a triangle to a finer resolution of development and experience periods.
        First it will attempt to disaggregate the triangle across experience periods putting
        different weights to field values over each subperiod. Next it will try to disaggregate
        across development axis applying interpolation.

        Implement this function with caution. It should not be used to refine transaction program
        data. Instead, it can be applied to triangles used for comparison or similar purposes
        during transaction analysis.

    Arguments:
        triangle: The triangle to disaggregate.
        resolution_exp_months: The resolution of experience periods in the output triangle.
        resolution_dev_months: The resolution of evaluation dates in the output triangle.
        fields: The list of fields in the triangle to disaggregate.
        period_weights: Weights for field values applied during disaggregation
            of one experience period. Weights must be from [0, 1], and should sum to 1.
            If set as a dictionary, each experience period will use a separate list of weights.
            If set as a list, the same weights are applied to all experience periods.
            If not supplied, weights are equally distributed across each experience period.
        interpolation_method: The interpolation method to use. See `scipy.interpolate.interp1d`
            to see the list of available interpolation methods.
        extrapolate_first_period: Boolean flag. If true, extrapolation will be applied to get
            field values for development periods prior to the first evaluation date.
        interpolation_kwargs: Keyword arguments passed to the interpolation method.
    Returns:
        A triangle with disaggregated experience periods and evaluation dates
            at the desired resolutions.
    """
    if min(triangle.dev_lags()) < 0:
        raise TriangleError(
            "Cannot disaggregate a triangle with negative development lags."
        )
    disaggregate_experience_triangle = disaggregate_experience(
        triangle,
        resolution_exp_months,
        period_weights,
        fields,
    )
    disaggregate_triangle = disaggregate_development(
        disaggregate_experience_triangle,
        resolution_dev_months,
        fields,
        interpolation_method,
        extrapolate_first_period,
        **interpolation_kwargs,
    )
    return disaggregate_triangle


def disaggregate_development(
    triangle: Triangle,
    resolution_months: int,
    fields: list[str] | None = None,
    interpolation_method: str = "linear",
    extrapolate_first_period: bool = True,
    **interpolation_kwargs,
) -> Triangle:
    """
    Disaggregate a triangle to a finer resolution of development periods
        using a selected interpolation method.

    Arguments:
        triangle: The triangle to disaggregate across development periods.
        resolution_months: The resolution of evaluation dates in the output triangle.
        fields: The list of fields in the triangle to interpolate.
        interpolation_method: The interpolation method to use. See `scipy.interpolate.interp1d`
            to see the list of available interpolation methods.
        extrapolate_first_period: Boolean flag. If true, extrapolation will be applied to get
            field values for development periods prior to the first evaluation date.
        interpolation_kwargs: Keyword arguments passed to the interpolation method.
    Returns:
        A triangle with interpolated evaluation dates at the desired resolution.
    """
    if resolution_months > eval_date_resolution(triangle):
        raise ValueError(
            f"Triangle has finer resolution for evaluation dates than {resolution_months=}."
        )
    elif resolution_months == eval_date_resolution(triangle):
        return triangle
    else:
        if fields is None:
            fields = DEFAULT_INTERPOLATION_FIELDS
        if not any([field in fields for field in triangle.fields]):
            raise ValueError(
                f"list {fields} should include at least one valid field from the input triangle."
            )

        cum_triangle = to_cumulative(triangle) if triangle.is_incremental else triangle
        cells = []
        for _, tri_slice in cum_triangle.slices.items():
            cells += _disaggregate_development_slice(
                tri_slice,
                resolution_months,
                fields,
                interpolation_method,
                extrapolate_first_period,
                **interpolation_kwargs,
            )
        disaggregate_triangle = (
            to_incremental(Triangle(cells))
            if triangle.is_incremental
            else Triangle(cells)
        )
        return disaggregate_triangle


def _disaggregate_development_slice(
    tri_slice: Triangle,
    resolution_months: int,
    fields: list[str],
    interpolation_method: str,
    extrapolate_first_period: bool,
    **interpolation_kwargs,
) -> list[Cell]:
    eval_resolution = eval_date_resolution(tri_slice)

    new_cells = []
    for _, row_cells in tri_slice.slice_period_rows:
        init_cell = row_cells[0]
        row_tri = Triangle(row_cells)

        dev_lags = row_tri.dev_lags(unit="months")
        # need at least 2 cells per period for interpolation
        if len(dev_lags) > 1:
            # check for gaps in development periods
            dev_offset = dev_lags[1] - dev_lags[0]
            for prv, nxt in zip(dev_lags[1:-1], dev_lags[2:]):
                if nxt - prv != dev_offset:
                    raise TriangleError(
                        "Cannot disaggregate a triangle with gaps in development periods."
                    )

            # cells in `row_cells` are already sorted by evaluation_date
            min_lag = init_cell.dev_lag()
            if extrapolate_first_period:
                min_x = np.maximum(0.0, min_lag - eval_resolution + resolution_months)
            else:
                min_x = min_lag
            max_x = row_cells[-1].dev_lag()
            pred_x = np.arange(min_x, max_x + resolution_months, resolution_months)

            field_values = {}
            for field in fields:
                if field not in row_tri.fields:
                    continue

                obs_x = np.array([cell.dev_lag() for cell in row_tri if field in cell])

                num_field_samples = row_tri.select([field]).num_samples
                if num_field_samples == 1:
                    obs_y = np.array([cell[field] for cell in row_tri if field in cell])
                else:
                    obs_y = np.array(
                        [
                            (
                                cell[field]
                                if isinstance(cell[field], np.ndarray)
                                and cell[field].size > 1
                                else np.full((num_field_samples,), cell[field])
                            )
                            for cell in row_tri
                            if field in cell
                        ]
                    ).T

                if extrapolate_first_period:
                    obs_x = np.concatenate([[-resolution_months], obs_x])
                    if num_field_samples == 1:
                        obs_y = np.concatenate([[0], obs_y])
                    else:
                        obs_y = np.concatenate(
                            [np.zeros(num_field_samples).reshape(-1, 1), obs_y], axis=1
                        )

                interp_fn = interp1d(
                    obs_x,
                    obs_y,
                    kind=interpolation_method,
                    fill_value="extrapolate" if extrapolate_first_period else None,
                    **interpolation_kwargs,
                )
                pred_y = interp_fn(pred_x)

                disaggregate_field_values = np.array(
                    [pred_y[..., ndx] for ndx in range(len(pred_x))]
                )
                if (disaggregate_field_values < 0).any():
                    raise ValueError(
                        "Negative field values are generated. Consider changing interpolation method."
                    )
                else:
                    field_values[field] = disaggregate_field_values

            new_cells += [
                init_cell.replace(
                    evaluation_date=add_months(init_cell.period_end, dev_lag),
                    values={field: field_values[field][ndx] for field in field_values},
                )
                for ndx, dev_lag in enumerate(pred_x)
            ]

        # now deal with period containing a single cell
        else:
            if not extrapolate_first_period:
                new_cells += [init_cell]
            else:
                if not new_cells:
                    continue
                cell_lag = init_cell.dev_lag()
                min_x = np.maximum(0.0, cell_lag - eval_resolution + resolution_months)
                pred_x = np.arange(
                    min_x, cell_lag + resolution_months, resolution_months
                )

                last_interpolated_period = sorted(new_cells, key=lambda c: c.period)[
                    -1
                ].period
                relevant_cells = sorted(
                    [
                        cell
                        for cell in new_cells
                        if cell.period == last_interpolated_period
                    ],
                    key=lambda c: c.dev_lag(),
                )[: len(pred_x)]

                new_cells += [
                    init_cell.replace(
                        evaluation_date=add_months(init_cell.period_end, dev_lag),
                        values={
                            field: cell[field]
                            * init_cell[field]
                            / relevant_cells[-1][field]
                            for field in fields
                            if field in cell.values
                        },
                    )
                    for dev_lag, cell in zip(pred_x, relevant_cells)
                ]

    return new_cells


def disaggregate_experience(
    triangle: Triangle,
    resolution_months: int,
    period_weights: Union[list[float], dict[date, list[float]]] | None = None,
    fields: list[str] | None = None,
) -> Triangle:
    """
    Disaggregate a triangle to a finer resolution across experience periods.
        The process allows to put different weights to each subinterval
        within one experience period.

    Arguments:
        triangle: The triangle to disaggregate across experience periods.
        resolution_months: The resolution of experience periods in the output triangle.
        period_weights: Weights for field values applied during disaggregation
            of one experience period. Weights must be from [0, 1], and should sum to 1.
            If set as a dictionary, each experience period will use a separate list of weights.
            If set as a list, the same weights are applied to all experience periods.
            If not supplied, weights are equally distributed across each experience period.
        fields: The list of fields in the triangle to disaggregate.
    Returns:
        A triangle with disaggregated field values at the desired resolution.
    """
    if not triangle.is_semi_regular():
        raise TriangleError("Triangle must be semi-regular.")

    tri_period_resolution = period_resolution(triangle)
    if resolution_months > tri_period_resolution:
        raise ValueError(
            f"Triangle has finer resolution for experience periods than {resolution_months=}."
        )
    elif resolution_months == tri_period_resolution:
        return triangle
    else:
        if fields is None:
            fields = DEFAULT_INTERPOLATION_FIELDS
        if not any([field in fields for field in triangle.fields]):
            raise ValueError(
                f"list {fields} should include at least one valid field from the input triangle."
            )

        if tri_period_resolution % resolution_months != 0:
            raise ValueError(
                f"Triangle experience periods cannot be disaggregated \
                over {resolution_months} months."
            )
        else:
            n_subperiods = tri_period_resolution // resolution_months
            if period_weights is None:
                period_weights = [1 / n_subperiods for i in range(0, n_subperiods)]
            _validate_period_weights(
                period_weights,
                n_subperiods,
                [period[0] for period in triangle.periods],
            )

        cum_triangle = to_cumulative(triangle) if triangle.is_incremental else triangle
        cells = []
        for _, tri_slice in cum_triangle.slices.items():
            cells += _disaggregate_experience_slice(
                tri_slice,
                resolution_months,
                period_weights,
                fields,
            )

        disaggregate_triangle = (
            to_incremental(Triangle(cells))
            if triangle.is_incremental
            else Triangle(cells)
        )
        return disaggregate_triangle


def _disaggregate_experience_slice(
    tri_slice: Triangle,
    resolution_months: int,
    period_weights: Union[list[float], dict[date, list[float]]],
    fields: list[str],
) -> list[Cell]:
    tri_period_resolution = period_resolution(tri_slice)
    n_periods = tri_period_resolution // resolution_months

    new_cells = []
    for cell in tri_slice.cells:
        subperiods = [
            (
                add_months(cell.period_start, k * resolution_months),
                add_months(cell.period_start, (k + 1) * resolution_months)
                - relativedelta(days=1),
            )
            for k in range(0, n_periods)
        ]
        # filter out periods after the cell evaluation date to account for negative dev lags
        subperiods = [
            period for period in subperiods if period[1] <= cell.evaluation_date
        ]
        cell_period_weights = period_weights[: len(subperiods)]
        sublist_total = sum(cell_period_weights)
        cell_period_weights = [weight / sublist_total for weight in cell_period_weights]

        weighted_values = _weight_cell_values(cell, cell_period_weights, subperiods)

        new_cells += [
            Cell(
                period_start=period[0],
                period_end=period[1],
                evaluation_date=cell.evaluation_date,
                metadata=cell.metadata,
                values={
                    field: weighted_values[period][field]
                    for field in cell.values
                    if field in fields
                },
            )
            for period in subperiods
            if period[1] <= cell.evaluation_date
        ]
    return new_cells


def _weight_cell_values(
    cell: Cell,
    period_weights: Union[list[float], dict[date, list[float]]],
    subperiods: list[Tuple[date, date]],
) -> dict[Tuple[date, date], dict[str, CellValue]]:
    period_values = dict()
    for field, value in cell.values.items():
        val = np.array([value]) if np.array(value).ndim == 0 else value
        weights = (
            np.array(period_weights[cell.period_start])
            if isinstance(period_weights, dict)
            else np.array(period_weights)
        )
        new_vals = np.array([v * weights for v in val])
        period_values[field] = {
            p: (v[0] if isinstance(value, (float, int)) else v)
            for p, v in zip(subperiods, new_vals.T)
        }
    # return data reorganized by subperiods
    return {
        key: {
            k: period_values[k][key] for k in period_values if key in period_values[k]
        }
        for key in subperiods
    }


def _validate_period_weights(
    period_weights: Union[list[float], dict[date, list[float]]],
    n_subperiods: int,
    tri_period_start: list[date],
):
    if isinstance(period_weights, list):
        weights = [period_weights]
    elif isinstance(period_weights, dict):
        weights = list(period_weights.values())
        if not all(ps in period_weights.keys() for ps in tri_period_start):
            raise ValueError(
                f"Must supply weights for all periods in {tri_period_start}."
            )
    else:
        raise TypeError(f"Set {period_weights=} as a list or dictionary.")

    if not all(len(w) == n_subperiods for w in weights):
        raise ValueError(
            f"All lists in variable {period_weights=} should have length {n_subperiods}."
        )
    if not all(0 <= val <= 1 for w in weights for val in w):
        raise ValueError(
            f"Each weight in {period_weights} must be in the range [0, 1]."
        )
    if not all(sum(w) == 1 for w in weights):
        raise ValueError(f"Weights in {period_weights} should sum to 1.")
    return
